/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * These .proto interfaces are private and stable.
 * Please see https://hadoop.apache.org/docs/current/hadoop-project-dist/hadoop-common/Compatibility.html
 * for what changes are allowed for a *stable* .proto interface.
 */
syntax = "proto2";
option java_package = "org.apache.hadoop.tracing";
option java_outer_classname = "TraceAdminPB";
option java_generic_services = true;
option java_generate_equals_and_hash = true;
package hadoop.common;

message ListSpanReceiversRequestProto {
}

message SpanReceiverListInfo {
  required int64 id = 1;
  required string className = 2;
}

message ListSpanReceiversResponseProto {
  repeated SpanReceiverListInfo descriptions = 1;
}

message ConfigPair {
  required string key = 1;
  required string value = 2;
}

message AddSpanReceiverRequestProto {
  required string className = 1;
  repeated ConfigPair config = 2;
}

message AddSpanReceiverResponseProto {
  required int64 id = 1;
}

message RemoveSpanReceiverRequestProto {
  required int64 id = 1;
}

message RemoveSpanReceiverResponseProto {
}

service TraceAdminService {
  rpc listSpanReceivers(ListSpanReceiversRequestProto)
      returns(ListSpanReceiversResponseProto);

  rpc addSpanReceiver(AddSpanReceiverRequestProto)
      returns(AddSpanReceiverResponseProto);

  rpc removeSpanReceiver(RemoveSpanReceiverRequestProto)
      returns(RemoveSpanReceiverResponseProto);
}
