#include <gssapi/gssapi.h>
#include <gssapi/gssapi_krb5.h>

const OM_uint32 _GSS_C_INDEFINITE = GSS_C_INDEFINITE;
const OM_uint32 _GSS_C_CALLING_ERROR_MASK = GSS_C_CALLING_ERROR_MASK;
const OM_uint32 _GSS_C_ROUTINE_ERROR_MASK = GSS_C_ROUTINE_ERROR_MASK;
const OM_uint32 _GSS_C_SUPPLEMENTARY_MASK = GSS_C_SUPPLEMENTARY_MASK;
const OM_uint32 _GSS_S_CALL_INACCESSIBLE_READ = GSS_S_CALL_INACCESSIBLE_READ;
const OM_uint32 _GSS_S_CALL_INACCESSIBLE_WRITE = GSS_S_CALL_INACCESSIBLE_WRITE;
const OM_uint32 _GSS_S_CALL_BAD_STRUCTURE = GSS_S_CALL_BAD_STRUCTURE;
const OM_uint32 _GSS_S_BAD_MECH = GSS_S_BAD_MECH;
const OM_uint32 _GSS_S_BAD_NAME = GSS_S_BAD_NAME;
const OM_uint32 _GSS_S_BAD_NAMETYPE = GSS_S_BAD_NAMETYPE;
const OM_uint32 _GSS_S_BAD_BINDINGS = GSS_S_BAD_BINDINGS;
const OM_uint32 _GSS_S_BAD_STATUS = GSS_S_BAD_STATUS;
const OM_uint32 _GSS_S_BAD_SIG = GSS_S_BAD_SIG;
const OM_uint32 _GSS_S_BAD_MIC = GSS_S_BAD_SIG;
const OM_uint32 _GSS_S_NO_CRED = GSS_S_NO_CRED;
const OM_uint32 _GSS_S_NO_CONTEXT = GSS_S_NO_CONTEXT;
const OM_uint32 _GSS_S_DEFECTIVE_TOKEN = GSS_S_DEFECTIVE_TOKEN;
const OM_uint32 _GSS_S_DEFECTIVE_CREDENTIAL = GSS_S_DEFECTIVE_CREDENTIAL;
const OM_uint32 _GSS_S_CREDENTIALS_EXPIRED = GSS_S_CREDENTIALS_EXPIRED;
const OM_uint32 _GSS_S_CONTEXT_EXPIRED = GSS_S_CONTEXT_EXPIRED;
const OM_uint32 _GSS_S_FAILURE = GSS_S_FAILURE;
const OM_uint32 _GSS_S_BAD_QOP = GSS_S_BAD_QOP;
const OM_uint32 _GSS_S_UNAUTHORIZED = GSS_S_UNAUTHORIZED;
const OM_uint32 _GSS_S_UNAVAILABLE = GSS_S_UNAVAILABLE;
const OM_uint32 _GSS_S_DUPLICATE_ELEMENT = GSS_S_DUPLICATE_ELEMENT;
const OM_uint32 _GSS_S_NAME_NOT_MN = GSS_S_NAME_NOT_MN;
const OM_uint32 _GSS_S_CONTINUE_NEEDED = GSS_S_CONTINUE_NEEDED;
const OM_uint32 _GSS_S_DUPLICATE_TOKEN = GSS_S_DUPLICATE_TOKEN;
const OM_uint32 _GSS_S_OLD_TOKEN = GSS_S_OLD_TOKEN;
const OM_uint32 _GSS_S_UNSEQ_TOKEN = GSS_S_UNSEQ_TOKEN;
const OM_uint32 _GSS_S_GAP_TOKEN = GSS_S_GAP_TOKEN;